/* 
   (c) Copyright 2001-2002  convergence integrated media GmbH.
   (c) Copyright 2002-2005  convergence GmbH.

   All rights reserved.

   Written by Denis Oliver Kropp <dok@directfb.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the
   Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

/*
 * @brief  This file contains definitions for the LiTE window interface.
 * @ingroup LiTE
 * @file font.h
 * <hr>
 */

#ifndef __LITE__WINDOW_H__
#define __LITE__WINDOW_H__

#ifdef __cplusplus
extern "C"
{
#endif

#include <direct/types.h>

#include <directfb.h>

#include "box.h"
#include "font.h"
#include "theme.h"

/* @brief Macro to convert a subclassed LiteWindow into a LiteWindow */
#define LITE_WINDOW(l) ((LiteWindow*)(l))

/* @brief LiteWindow blend modes */
typedef enum {
     LITE_BLEND_ALWAYS,         /**< Always blend */
     LITE_BLEND_NEVER,          /**< Never blend */
     LITE_BLEND_AUTO            /**< Automatically blend */
} LiteBlendMode;

/* @brief Window flags */
typedef enum {
    LITE_WINDOW_MODAL          = 1 << 1,        /**< Modal window */
    LITE_WINDOW_RESIZE         = 1 << 2,        /**< Resizable window */
    LITE_WINDOW_MINIMIZE       = 1 << 3         /**< Window that could be minimized */
} LiteWindowFlags;

/* @brief Window alignment flags */
typedef enum {
     LITE_CENTER_HORIZONTALLY = -1,     /**< Center window on the x-axis */
     LITE_CENTER_VERTICALLY = -2        /**< Center window on the y-axis */
}LiteAlignmentFlags;

/* @brief Full window opacity level (no alpha blending) */
#define liteFullWindowOpacity 0xff

/* @brief No window opacity (window invisible ) */
#define liteNoWindowOpacity 0x00

/* @brief LiteWindowTheme window theme structure */
typedef struct  _LiteWindowTheme {
     LiteTheme theme;                   /**< Base LiTE Theme */

     LiteFont *title_font;              /**< Title font */

     struct {
          struct {
               IDirectFBSurface *surface;       
               int               width;
               int               height;
          } left;                       /**< left surface */
          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } right;
          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } top;                        /**< top surface */
          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } bottom;                     /**< bottom surface */

          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } topleft;                    /**< top left surface */
          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } topright;                   /**< top right surface */
          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } bottomleft;                 /**< bottom left surface */
          struct {
               IDirectFBSurface *surface;
               int               width;
               int               height;
          } bottomright;                /**< bottom right surface */
     } frame;                           /**< Frame structure */
} LiteWindowTheme;

/* @brief Standard window themes */
/* @brief No window theme */
#define          liteNoWindowTheme  NULL

/* @brief LiTE default window theme */
extern LiteWindowTheme *liteDefaultWindowTheme;



/* @brief Window mouse event callback.
 * This prototype is used for defining a callback for cases 
 * where mouse event handling inside the window is intercepted.
 *
 * @param evt                   OUT:    Window event
 * @param data                  OUT:    Context data
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
typedef DFBResult (*WindowMouseFunc) (DFBWindowEvent* evt, void *data);

/* @brief Window keyboard event callback.
 * This prototype is used for defining a callback for cases 
 * where keyboard event handling inside the window is intercepted.
 *
 * @param evt                   OUT:    Window event
 * @param data                  OUT:    Context data
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
typedef DFBResult (*WindowKeyboardFunc) (DFBWindowEvent* evt, void *data);

/* @brief Window scroll wheel event callback.
 * This prototype is used for defining a callback for cases 
 * where mouse scroll event handling inside the window is intercepted.
 *
 * @param evt                   OUT:    Window event
 * @param data                  OUT:    Context data
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
typedef DFBResult (*WindowWheelFunc) (DFBWindowEvent* evt, void *data);

/* @brief Window event callback.
 * This prototype is used for defining a callback for cases 
 * where window events are intercepted before the toolbox has a 
 * chance to process these.
 *
 * @param evt                   OUT:    Window event
 * @param data                  OUT:    Context data
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 *                              Return any other event to indicate that
 *                              the event is processed and should not be
 *                              processed by the toolbox.
 */
typedef DFBResult (*WindowEventFunc) (DFBWindowEvent* evt, void *data);

/* @brief univesal event callback.
 * This prototype is used for defining a callback for cases 
 * where universal events inside the window is intercepted.
 *
 * @param evt                   OUT:    Universal event
 * @param data                  OUT:    Context data
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
typedef DFBResult (*UniversalEventFunc) (DFBUniversalEvent* evt, void *data);

/* @brief Window user event callback.
 * This prototype is used for defining a callback for cases 
 * where user events inside the window is intercepted.
 *
 * @param evt                   OUT:    User event
 * @param data                  OUT:    Context data
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
typedef DFBResult (*WindowUserEventFunc) (DFBUserEvent* evt, void *data);

/* @brief LiteWindow structure */
typedef struct _LiteWindow {
  LiteBox           box;                /**< Underlying LiteBox */
  struct _LiteWindow *creator;          /**< Creator window */

  int               width;              /**< Window width */
  int               height;             /**< Window height */

  DFBWindowID       id;                 /**< Window ID */

  char             *title;              /**< Window title */
  IDirectFBWindow  *window;             /**< Underlying DirectFB window */
  IDirectFBSurface *surface;            /**< Underlying DirectFB surface */

  LiteWindowFlags   flags;              /**< Window flags */
  int               moving;             /**< If the window could be moved or not */
  int               resizing;           /**< If the window could be resized or not */
  int               old_x;              /**< Original x position */
  int               old_y;              /**< Original y position */

  int               step_x;             /**< X steps */
  int               step_y;             /**< Y steps */

  int               min_width;          /**< Minimum width */
  int               min_height;         /**< Minimum height */

  int               last_width;         /**< Last width */
  int               last_height;        /**< Last height */

  DFBWindowEvent    last_resize;        /**< Last resize event */
  DFBWindowEvent    last_motion;        /**< Last motion event */

  struct timeval    last_click;         /**< Last time window was clicked */

  int               has_focus;          /**< Window has focus or not */

  WindowMouseFunc       raw_mouse_func;         /**< Raw mouse callback */
  void                 *raw_mouse_data;         /**< Raw mouse callback data*/
  WindowMouseFunc       mouse_func;             /**< Mouse callback */
  void                 *mouse_data;             /**< Mouse callback data */

  WindowKeyboardFunc    raw_keyboard_func;      /**< Raw keyboard callback */
  void                 *raw_keyboard_data;      /**< Raw keyboard callback data */
  WindowKeyboardFunc    keyboard_func;          /**< Keyboard callback */
  void                 *keyboard_data;          /**< Keyboard callback data */

  WindowEventFunc       window_event_func;      /**< Window event callback */
  void                 *window_event_data;      /**< Window event callback data */

  UniversalEventFunc    universal_event_func;   /**< Universal event callback */
  void                 *universal_event_data;   /**< Universal event callback data */

  WindowUserEventFunc   user_event_func;        /**< User event callback */
  void                 *user_event_data;        /**< User event callback data */

  WindowWheelFunc       wheel_func;             /**< Scroll wheel callback */
  void                 *wheel_data;             /**< Scroll wheel callback data */

  struct {
       DFBRegion    region;
       DFBBoolean   pending;
  } update;                             /**< Update area */

  LiteBlendMode     content_mode;       /**< Content blend mode */
  LiteBlendMode     opacity_mode;       /**< Opacity blend mode */

  struct {
       DFBBoolean   enabled;
       DFBColor     color;
  } bg;                                 /**< Background color*/

  LiteBox          *entered_box;        /**< Currently entered box */
  LiteBox          *focused_box;        /**< Currently focused box */
  LiteBox          *drag_box;           /**< Currently dragged box */

  LiteWindowTheme  *theme;              /**< Window theme installed */

  int (*OnMove)     (struct _LiteWindow *self, int x, int y);   /**< Move callback */
  int (*OnResize)   (struct _LiteWindow *self, int width, int height); /**< Resize callback */
  int (*OnClose)    (struct _LiteWindow *self);                 /**< Close callback */
  int (*OnDestroy)  (struct _LiteWindow *self);                 /**< Destroy callback */
  int (*OnFocusIn)  (struct _LiteWindow *self);                 /**< Focus in callback */
  int (*OnFocusOut) (struct _LiteWindow *self);                 /**< Focus out callback */
  int (*OnEnter)    (struct _LiteWindow *self, int x, int y);   /**< Enter callback */
  int (*OnLeave)    (struct _LiteWindow *self, int x, int y);   /**< Leave callback */
  int (*OnBoxAdded)   (struct _LiteWindow *self, LiteBox *box); /**< Box added callback */
  int (*OnBoxToBeRemoved) (struct _LiteWindow *self, LiteBox *box); /**< Box To be Removed callback */
} LiteWindow;

/* @brief Create a new LiteWindow object
 * This function will create a new LiteWindow object. You
 * could pass in the size, layer used, various capabilities
 * for the window (such as alpha blending), and a top part title 
 * in case the window has a window theme installed.
 * You could install the default theme (liteDefaultWindowTheme),
 * no theme (liteNoWindowTheme), or your own custom window theme
 * based on the LiteWindowTheme structure.
 *
 * @param layer                 IN:     Window display layer, if NULL
                                        the default layer is used.
 * @param rect                  IN:     Window size
 * @param cap                   IN:     Window capabilities.
 * @param theme                 IN:     Window theme
 * @param title                 IN:     Top part window title 
 * @param ret_window            OUT:    Valid window object
 * 
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_new_window(IDirectFBDisplayLayer *layer,
                          DFBRectangle          *rect,
                          DFBWindowCapabilities  caps,
                          LiteWindowTheme       *theme,
                          const char            *title,
                          LiteWindow           **ret_window);

/* @brief Initialize a Window object
 * This function will initialize a window object. This object
 * is mostly called from specific windows subclassed from the
 * window objeect. lite_new_window() will call this function.
 *
 * @param win                   IN:     Window object
 * @param layer                 IN:     Layer used
 * @param win_rect              IN:     Window size
 * @param caps                  IN:     Window capabilities
 * @param theme                 IN:     Window theme
 * @param title                 IN:     Window title          
 * 
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_init_window(LiteWindow            *win,
                           IDirectFBDisplayLayer *layer,
                           DFBRectangle          *win_rect,
                           DFBWindowCapabilities  caps,
                           LiteWindowTheme       *theme,
                           const char            *title);

/* @brief Set the window creator of a specific window
 * This function will store the original window that created a specific window.
 * This feature is needed in cases of modal windows that will open another LiteWindow,
 * if the window knows of the creator window, it will forward all events to the new
 * window created. In practice it means that all input events are forwarded to a new window
 * that is bigger than the original modal window.

 * @param window                IN:     The window that needs to know about its creator window
 * @param creator               IN:     The window that is creating the new LiteWindow.
 *
 * @return DFBResult            Returns DFB_OK if successful.
 */
DFBResult lite_window_set_creator(LiteWindow *window, LiteWindow *creator);

/* @brief Get the window creator of a specific window
 * This function will get the creator window of a specific window, or the
 * window that created a specific window.  
 * It is assumed that an earlier lite_window_set_creator() has been called, if
 * not NULL will be returned.
 *
 * @param window                IN:     The window that needs to know about its creator window
 * @param creator               OUT:    The window that created a specific LiteWindow
 *
 * @return DFBResult            Returns DFB_OK if successful.
 */
DFBResult lite_window_get_creator(LiteWindow *window, LiteWindow **creator);

/* @brief Set the modal state of a window
 * This function will set the modal state of a window, if enabled
 * the window will be only window receiving events. If false
 * all created windows will receive events.
 *
 * @param win                   IN:     Valid window object
 * @param modal                 IN:     Modal state of the window
 * 
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_window_set_modal(LiteWindow *window, bool modal);
           
/* @brief Start the window event loop
 * This function start the event loop. You only need to do this
 * for one (main) window, all other windows will equally receive
 * events unless a window has been set into modal state.
 * If you pass in a timeout value the event loop will run until
 * the timeout has expired, and then return from the function. This
 * could be used for creating idle event handling in the actual
 * application.
 *
 * @param win                   IN:     Valid window object
 * @param timeout               IN:     Timeout value for the event loop
 * 
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_window_event_loop(LiteWindow *window, int timeout);

/* @brief Update the window
 * This function will update the window in the specified region.
 * If NULL is passed as a region the whole window area will be updated.
 *
 * @param win                   IN:     Valid window object
 * @param region                IN:     Window region
 * 
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_update_window(LiteWindow *window, const DFBRegion *region);

/* @brief Set the window title
 * This function will set the window title provided the window
 * has a theme installed.
 *
 * @param win                   IN:     Valid window object
 * @param title                 IN:     Window title 
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_set_window_title(LiteWindow *window, const char *title);

/* @brief Set the window opacity level
 * This function will set the window opacity level.
 * Note that after creating a window the opacity level is 0 (liteNoWindowOpacity),
 * so you need to set it to a higher value, such as liteFullWindowOpacity (0xff) 
 * in order for the window to show up.
 *
 * @param win                   IN:     Valid window object
 * @param opacity               IN:     Opacity level, 0 to 0xff, 
 *                                      liteNoWindowOpacity corresponds to 0,
 *                                      liteFullWindowOpacity corresponds to 0xff 
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_set_window_opacity(LiteWindow *window, __u8 opacity);

/* @brief Set the window background color
 * This function will set the window background color. Note
 * that each window has a default window color set in the
 * lite configuration file.
 *
 * @param win                   IN:     Valid window object
 * @param color                 IN:     Color specification
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_set_window_background(LiteWindow *window, DFBColor *color);

/* @brief Set the window background color
 * This function will set the window background color. Note
 * that each window has a default window color set in the
 * lite configuration file.
 *
 * @param win                   IN:     Valid window object
 * @param r                     IN:     R color component
 * @param g                     IN:     G color component
 * @param b                     IN:     B color component
 * @param a                     IN:     Alpha component. 
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_set_window_background_color(LiteWindow *window,
                                           __u8        r,
                                           __u8        g,
                                           __u8        b,
                                           __u8        a);

/* @brief Set the window blend mode
 * This function will set the window blend mode.
 *
 * @param win                   IN:     Valid window object
 * @param content               IN:     Window content blend mode
 * @param opacity               IN:     Opacity blend mode
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_set_window_blend_mode(LiteWindow   *window,
                                     LiteBlendMode content,
                                     LiteBlendMode opacity);

/* @brief Resize the window
 * This function will resize the window if the resize flag is 
 * enabled. By default the resize flag is set.
 *
 * @param win                   IN:     Valid window object
 * @param width                 IN:     New window width
 * @param height                IN:     New window height
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_resize_window(LiteWindow  *window,
                             unsigned int width,
                             unsigned int height);

/* @brief Get Window size
 * This function will get the current window size.
 *
 * @param win                   IN:     Valid window object
 * @param width                 OUT:    Window width
 * @param height                OUT:    Window height
 *
 * @return DFBResult            Returns DFB_OK if successful.
 *
 */
DFBResult lite_get_window_size(LiteWindow *window, unsigned int *width,
                                unsigned int *height);

/* @brief Minimize the window
 * This function will minimze the window if the minimize flag 
 * is set. By default the minimize flag is set.
 *
 * @param win                   IN:     Valid window object
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_minimize_window(LiteWindow *window);

/* @brief Maximize the window
 * This function will maximize the window if the minimize flag 
 * is set and the window is minimized. By default the minimize flag is set.
 *
 * @param win                   IN:     Valid window object
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_restore_window(LiteWindow *window);

/* @brief Set focus to a specific LiteBox
 * This function will set the focus flag on a specified LiteBox in
 * the window.
 *
 * @param box                   IN:     LiteBox to be focused
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_focus_box(LiteBox *box);

/* @brief Handle window events
 * This function will call the function to handle incoming window
 * events. It is called by default from the window event loop, and
 * is usually not needed unless you want to directly call window event
 * handling from external code.
 * 
 * @param win                   IN:     Valid window object
 * @param event                 IN:     Valid window event
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
int lite_handle_window_event(LiteWindow *window, DFBWindowEvent *event);

/* @brief Flush window events
 * This function will flush any outstanding window events.
 * 
 * @param win                   IN:     Valid window object
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_flush_window_events(LiteWindow *window);

/* @brief Find window that a LiteBox belongs to
 * This function will find what window a specific LiteBox belongs to.
 * 
 * @param box                   IN:     Valid LiteBox
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
LiteWindow *lite_find_my_window(LiteBox *box);

/* @brief Close window
 * This function will close a window. If this window is the last
 * window in the application, it will exit the event loop.
 * 
 * @param win                   IN:     Valid window object
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_close_window(LiteWindow *window);

/* @brief Destroy window
 * This function will close a window and destroy the window resources. 
 * If this window is the last window in the application, 
 * it will exit the event loop.
 * 
 * @param win                   IN:     Valid window object
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_destroy_window(LiteWindow *window);

/* @brief Install a raw mouse event callback
 * This function will install a raw mouse event callback that
 * is triggered before the window has processed the mouse event
 * and delegated it to a specific LiteBox. If DFB_OK is not returned
 * then the event handling for this mouse event is terminated.
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_raw_window_mouse(LiteWindow     *window,
                                   WindowMouseFunc callback,
                                   void           *data);

/* @brief Install a mouse event callback
 * This function will install a raw mouse event callback that
 * is triggered after the window has processed the mouse event
 * and delegated it to a specific LiteBox. 
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_window_mouse(LiteWindow     *window, 
                               WindowMouseFunc callback, 
                               void           *data);

/* @brief Install a raw keyboard event callback
 * This function will install a raw keyboard event callback that
 * is triggered before the window has processed the keyboard event
 * and delegated it to a specific LiteBox. If DFB_OK is not returned
 * then the event handling for this keyboard event is terminated.
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_raw_window_keyboard(LiteWindow        *window, 
                                      WindowKeyboardFunc callback, 
                                      void              *data);

/* @brief Install a keyboard event callback
 * This function will install a keyboard event callback that
 * is triggered after the window has processed the keyboard event
 * and delegated it to a specific LiteBox.
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_window_keyboard(LiteWindow        *window, 
                                  WindowKeyboardFunc callback, 
                                  void              *data);

/* @brief Install a window event callback
 * This function will install a window event callback that
 * is triggered before the LiTE toolbox has processed the actual event.
 * This could be used to filter or otherwise override any default window
 * event processing. If the callback installs does return a value not
 * equal to DFB_OK, then it's assumed that this event should not be forwarded
 * to the toolbox.
 * 
 * This function should be carefully used, as the callback will override all
 * default behavior of the toolbox.
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_window_event(LiteWindow     *window,
                               WindowEventFunc callback,
                               void           *data);

/* @brief Install a universal event callback
 * This function will install a universal event callback that
 * is triggered when the window is processing universal events
 * Note that LiTE is internally not using any universal events.
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_window_universal_event(LiteWindow         *window,
                                         UniversalEventFunc  callback,
                                         void               *data);

/* @brief Install a user event callback
 * This function will install a user event callback that
 * is triggered when the window is processing custom user events
 * Note that LiTE is internally not using any user events.
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_window_user_event(LiteWindow         *window,
                                    WindowUserEventFunc callback,
                                    void               *data);

/* @brief Install a scroll wheel event callback
 * This function will install a scroll wheel event callback that
 * is triggered when the window is processing scroll wheel events
 * 
 * @param window                IN:     Valid window object
 * @param callback              IN:     Callback to be triggered
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_on_window_wheel(LiteWindow     *window,
                               WindowWheelFunc callback,
                               void           *data);

/* @brief Post custom events
 * This function will post events to the event queue.
 * Note that you could post any kinds of events to the queue,
 * including window events and user events.
 * 
 * @param window                IN:     Valid window object
 * @param event                 IN:     Custom event
 * @param data                  IN:     Data context
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_post_event_to_window(LiteWindow *window, DFBEvent *event);

/* @brief Release window resources 
 * This function will release all window resources.
 * 
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_release_window_resources (void);

/* @brief Get access to the event queue
 * This function will return the event queue pointer.
 * This could be used in case you need to post or peek inside
 * the event queue outside the window code itself.
 * 
 * @param ret_event_buffer      OUT:    Current event buffer
 *
 * @return DFBResult            Returns DFB_OK if successful.     
 */
DFBResult lite_get_event_buffer(IDirectFBEventBuffer **ret_event_buffer);

/* @brief Free the window theme resources
 *
 * This function will release the window theme resources
 * including freeing the used memory in the various
 * sub-surfaces.
 *
 * @param theme                 IN:     Window theme
 *
 * @return DFBResult            Returns DFB_OK if successful.
 */
DFBResult
lite_free_window_theme(LiteWindowTheme *theme);



#ifdef __cplusplus
}
#endif

#endif /*  __LITE__WINDOW_H__  */
